# AUTOGENERATED! DO NOT EDIT! File to edit: ../../../nbs/agents/processors/01_processors.ipynb.

# %% auto 0
__all__ = ['RoundAction', 'ClipNegative', 'DiscreteToContinuous', 'OneHotToDiscrete', 'HybridToContinuous',
           'OneHotHybridtoContinuous', 'ContToBinary', 'MakeInputZero', 'GetTimeSeriesAndStaticFeatures']

# %% ../../../nbs/agents/processors/01_processors.ipynb 4
# General libraries:
import numpy as np

# %% ../../../nbs/agents/processors/01_processors.ipynb 6
class RoundAction():
    """
    This class implements a processor that rounds the action to the closest
    multiple of a given number.

    args:
        increment (float or np.array): the increment to be used for rounding.

    """
    

    def __init__(self, increment):
        self.increment = increment
    
    def __call__(self, input):
        return np.round(input / self.increment) * self.increment

# %% ../../../nbs/agents/processors/01_processors.ipynb 7
class ClipNegative():
    """
    This class implements a processor that clips the input to zero if it is
    negative.
    
    """

    def __call__(self, input):
        return input.copy().clip(min=0)

# %% ../../../nbs/agents/processors/01_processors.ipynb 8
class DiscreteToContinuous():
    """
    This class implements a processor that converts a discrete action to a
    continuous action. The discrete action is assumed to be an integer between
    0 and n_actions - 1, and the continuous action is assumed to be between
    action_min and action_max.

    args:
        n_actions (int): the number of actions.
        action_min (np.array): the minimum value of each continuous action.
        action_max (np.array): the maximum value of each continuous action.

    """

    def __init__(self, n_actions, action_min, action_max):
        if isinstance(n_actions, int):
            n_actions = np.array([n_actions])
        assert np.all(n_actions%2 != 0) , "Need odd number of actions to ensure 0 is included"
        self.n_actions = n_actions
        self.action_min = action_min
        self.action_max = action_max
    
    def __call__(self, action):
        action = action * (self.action_max - self.action_min) / (self.n_actions - 1) + self.action_min
        return action

# %% ../../../nbs/agents/processors/01_processors.ipynb 9
class OneHotToDiscrete():

    """
    This class implements a processor that converts a one-hot action to a
    discrete action. The one-hot action is assumed to be a one-hot vector of
    length n_actions, and the discrete action is assumed to be an integer
    between 0 and n_actions - 1.

    args:
        n_actions (int): the number of actions.

    """
    # TODO: make this work for multidimensional one-hot vectors

    def __init__(self):
        pass
    
    def __call__(self, action):
        action = action.argmax(axis=-1)
        return action

# %% ../../../nbs/agents/processors/01_processors.ipynb 10
class HybridToContinuous():

    """

    This class implements a processor that converts a hybrid action to a
    continuous action. The hybrid action is assumed to be a vector of length
    n_actions * 2 where the first half corresponds to the decision of whether
    to order or not, and the second half corresponds to the quantity.

    args:
        n_actions (int): the number of actions.

    """

    def __init__(self, num_products):
        self.num_products = num_products
    
    def __call__(self, action):
        order_action = action[:self.num_products]
        order_quantity = action[self.num_products:]
        order_action = np.where(order_action > 0, 1, 0)
        action = order_action * order_quantity
        return action

# %% ../../../nbs/agents/processors/01_processors.ipynb 11
class OneHotHybridtoContinuous():

    """

    This class implements a processor that converts a hybrid action that has a one-
    hot vector for the discrete action into the hybrid action that has a scalar value
    for the discrete action

    args:
        n_actions (int): the number of actions.

    """

    def __init__(self, num_products):
        self.num_products = num_products
    
    def __call__(self, action):
        # TODO check if this works for more than 1 product
        order_quantity = action[:self.num_products]
        order_action = action[self.num_products:]
        for i in range(self.num_products):
            order_action_prod = order_action[i:i+2]
            order_action = np.argmax(order_action_prod)
        action = order_action * order_quantity
        return action

# %% ../../../nbs/agents/processors/01_processors.ipynb 12
class ContToBinary():

    """
    This class implements a processor that converts a continuous input 
    into a binary input. The binary feature will be one if the continuous
    input is greater than 0, and zero otherwise.

    #! Currently only works for single product

    """

    def __init__(self):
        pass

    def __call__(self, state):
        return np.array([1]) if state[0] > 0 else np.array([0])

# %% ../../../nbs/agents/processors/01_processors.ipynb 13
class MakeInputZero():
    
        """
        This class implements a processor that makes the input zero.
    
        """
    
        def __init__(self):
            pass
    
        def __call__(self, state):
            return np.zeros_like(state)

# %% ../../../nbs/agents/processors/01_processors.ipynb 14
class GetTimeSeriesAndStaticFeatures():
    """
    This class implements a processor that splits the state into time series and static features.
    The time series features are assumed to be the first num_ts_features * len_ts_features features
    and the static features are assumed to be the remaining features.

    args:
        num_ts_features (int): the number of time series features.
        len_ts_features (int): the length of each time series feature.

    """

    def __init__(self, num_ts_features, len_ts_features):
        self.num_ts_features = num_ts_features
        self.len_ts_features = len_ts_features
        
    def __call__(self, state):

        ts_data = state[:,:self.num_ts_features*self.len_ts_features]
        static_data = state[:,self.num_ts_features*self.len_ts_features:]

        ts_data = ts_data.reshape(state.shape[0], self.num_ts_features, self.len_ts_features).transpose(1, 2)

        return [ts_data, static_data]



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../../nbs/agents/benchmark_agents/02_SAA_agents.ipynb.

# %% auto 0
__all__ = ['FakePolicy', 'WSAAAgent', 'check_cu_co', 'SAAAgent']

# %% ../../../nbs/agents/benchmark_agents/02_SAA_agents.ipynb 4
# General libraries:
import numpy as np
from scipy.stats import norm
import numbers

from sklearn.ensemble import RandomForestRegressor
from sklearn.utils.validation import check_array
import pulp

# Mushroom libraries
from mushroom_rl.core import Agent

# %% ../../../nbs/agents/benchmark_agents/02_SAA_agents.ipynb 7
class FakePolicy():
    def reset():
        pass

class WSAAAgent(Agent):
    def __init__(self,
                 cu=None,
                 co=None,
                 criterion="mse",
                 n_estimators=100,
                 max_depth=None,
                 min_samples_split=6, # starndard 2
                 min_samples_leaf=3, # standard 1
                 min_weight_fraction_leaf=0.,
                 max_features="auto",
                 max_leaf_nodes=None,
                 min_impurity_decrease=0.,
                 bootstrap=True,
                 oob_score=False,
                 n_jobs=None,
                 random_state=None,
                 verbose=0,
                 warm_start=False,
                 ccp_alpha=0.0,
                 max_samples=None,
                 weight_function="w1",
                 agent_name = "wSAA"
                 ):
        self.criterion = criterion
        self.n_estimators = n_estimators
        self.max_depth = max_depth
        self.min_samples_split = min_samples_split
        self.min_samples_leaf = min_samples_leaf
        self.min_weight_fraction_leaf = min_weight_fraction_leaf
        self.max_features = max_features
        self.max_leaf_nodes = max_leaf_nodes
        self.min_impurity_decrease = min_impurity_decrease
        self.bootstrap = bootstrap
        self.oob_score = oob_score
        self.n_jobs = n_jobs
        self.random_state = random_state
        self.verbose = verbose
        self.warm_start = warm_start
        self.ccp_alpha = ccp_alpha
        self.max_samples = max_samples
        self.weight_function = weight_function
        self.cu=cu
        self.co=co

        self.train_directly=True
        self.train_mode = "direct"
        self.policy = FakePolicy

        self._postprocessors=list()
        self._preprocessors=list() 

        self.name = agent_name
        self.fitted=False
        
    def _get_fitted_model(self, X, y):
        model = RandomForestRegressor(
            criterion=self.criterion,
            n_estimators=self.n_estimators,
            max_depth=self.max_depth,
            min_samples_split=self.min_samples_split,
            min_samples_leaf=self.min_samples_leaf,
            min_weight_fraction_leaf=self.min_weight_fraction_leaf,
            max_features=self.max_features,
            max_leaf_nodes=self.max_leaf_nodes,
            min_impurity_decrease=self.min_impurity_decrease,
            bootstrap=self.bootstrap,
            oob_score=self.oob_score,
            n_jobs=self.n_jobs,
            random_state=self.random_state,
            verbose=self.verbose,
            warm_start=self.warm_start,
            ccp_alpha=self.ccp_alpha,
            max_samples=self.max_samples
        )

        self.model_ = model.fit(X, y)
        self.train_leaf_indices_ = model.apply(X)
        

    def _calc_weights(self, sample):
        sample_leaf_indices = self.model_.apply([sample])
        if self.weight_function == "w1":
            n = np.sum(sample_leaf_indices == self.train_leaf_indices_, axis=0)
            treeWeights = (sample_leaf_indices == self.train_leaf_indices_) / n
            weights = np.sum(treeWeights, axis=1) / self.n_estimators
        else:
            n = np.sum(sample_leaf_indices == self.train_leaf_indices_)
            treeWeights = (sample_leaf_indices == self.train_leaf_indices_) / n
            weights = np.sum(treeWeights, axis=1)
        
        weightPosIndex = np.where(weights > 0)[0]
        weightsPos = weights[weightPosIndex]

        return (weightsPos, weightPosIndex)
    

    def fit(self, features, demand):

        X=features
        y=demand

        cu = self.cu
        co = self.co

        self._get_fitted_model(X, y)

        if y.ndim == 1:
            y = np.reshape(y, (-1, 1))

        # Training data
        self.y_ = y
        self.X_ = X
        self.n_samples_ = y.shape[0]

        # Determine output settings
        self.n_outputs_ = y.shape[1]
        self.n_features_ = X.shape[1]

        self.cu = cu
        self.co = co

        self.fitted=True

        self.model_.verbose=0

        return self

    def _findQ(self, weights, weightPosIndices):
        
        y = self.y_
        yWeightPos = y[weightPosIndices]
        
        q = []
        
        for i in range(self.n_outputs_):
            serviceLevel = self.cu[i] / (self.cu[i] + self.co[i])
            
            indicesYSort = np.argsort(yWeightPos[:, i])
            ySorted = yWeightPos[indicesYSort, i]
            
            distributionFunction = np.cumsum(weights[indicesYSort])
            decisionIndex = np.where(distributionFunction >= serviceLevel)[0][0]
            
            q.append(ySorted[decisionIndex])
        
        return q

    def draw_action(self, X):   

        if self.fitted:  

            if X.ndim == 1:
                X = np.reshape(X, (-1, self.n_features_))
            
            weightsDataList = [self._calc_weights(row) for row in X]

            pred = [self._findQ(weights, weightPosIndices) 
                    for weights, weightPosIndices in weightsDataList]
            pred = np.array(pred).squeeze()

        else:
            pred = np.random.rand(1)  
        
        return pred


# %% ../../../nbs/agents/benchmark_agents/02_SAA_agents.ipynb 8
def check_cu_co(cu, co, n_outputs):
    """Validate under- and overage costs.

    Parameters
    ----------
    cu : {ndarray, Number or None}, shape (n_outputs,)
       The underage costs per unit. Passing cu=None will output an array of ones.
    co : {ndarray, Number or None}, shape (n_outputs,)
       The overage costs per unit. Passing co=None will output an array of ones.
    n_outputs : int
       The number of outputs.
    Returns
    -------
    cu : ndarray, shape (n_outputs,)
       Validated underage costs. It is guaranteed to be "C" contiguous.
    co : ndarray, shape (n_outputs,)
       Validated overage costs. It is guaranteed to be "C" contiguous.
    """
    costs = [[cu, "cu"], [co, "co"]]
    costs_validated = []
    for c in costs:
        if c[0] is None:
            cost = np.ones(n_outputs, dtype=np.float64)
        elif isinstance(c[0], numbers.Number):
            cost = np.full(n_outputs, c[0], dtype=np.float64)
        else:
            cost = check_array(
                c[0], accept_sparse=False, ensure_2d=False, dtype=np.float64,
                order="C"
            )
            if cost.ndim != 1:
                raise ValueError(c[1], "must be 1D array or scalar")

            if cost.shape != (n_outputs,):
                raise ValueError("{}.shape == {}, expected {}!"
                                 .format(c[1], cost.shape, (n_outputs,)))
        costs_validated.append(cost)
    cu = costs_validated[0]
    co = costs_validated[1]
    return cu, co

class SAAAgent(Agent):

    #BaseNewsvendor, ClassicMixin

    def __init__(self,
                 cu=None,
                 co=None,
                 agent_name="SAA"
                 ):
        self.cu = cu
        self.co = co

        self.train_directly=True
        self.train_mode = "direct"
        self.policy = FakePolicy

        self._postprocessors=list()
        self._preprocessors=list() 

        self.name = agent_name
        self.fitted=False

    def _calc_weights(self):
        weights = np.full(self.n_samples_, 1 / self.n_samples_)
        return weights

    def fit(self, demand, features):

        y=demand
       
        y = check_array(y, ensure_2d=False, accept_sparse='csr')

        if y.ndim == 1:
            y = np.reshape(y, (-1, 1))

        # Training data
        self.y_ = y
        self.n_samples_ = y.shape[0]

        # Determine output settings
        self.n_outputs_ = y.shape[1]

        # Check and format under- and overage costs
        self.cu_, self.co_ = check_cu_co(self.cu, self.co, self.n_outputs_)

        self.q_star = np.array(self._findQ(self._calc_weights()))

        self.fitted=True

        return self

    def _findQ(self, weights):
        """Calculate the optimal order quantity q"""

        y = self.y_
        q = []

        for k in range(self.n_outputs_):
            alpha = self.cu_[k] / (self.cu_[k] + self.co_[k])
            q.append(np.quantile(y[:, k], alpha, interpolation="higher"))

        return q

    def draw_action(self, *args, **kwargs):

        if self.fitted:
            pred = self.q_star
            
        else:
            pred = np.random.rand(1)  
        return pred



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../../nbs/agents/networks/00_base.ipynb.

# %% auto 0
__all__ = ['BaseNetwork']

# %% ../../../nbs/agents/networks/00_base.ipynb 4
# General libraries
import numpy as np

# Pytorch
import torch
import torch.nn as nn
import torch.nn.functional as F

# Other libraries
from abc import ABC, abstractmethod

# %% ../../../nbs/agents/networks/00_base.ipynb 6
class BaseNetwork(nn.Module, ABC):
    """
    Base class for simple 3-Layer network.
    """

    # TODO make the network architecture more flexible

    def __init__(self, input_shape, output_shape, n_features, **kwargs):
        super().__init__()

        n_input = input_shape[-1]
        n_output = output_shape[0]

        self._h1 = nn.Linear(n_input, n_features)
        self._h2 = nn.Linear(n_features, n_features)
        self._h3 = nn.Linear(n_features, n_output)

        nn.init.xavier_uniform_(self._h1.weight,
                                gain=nn.init.calculate_gain('relu'))
        nn.init.xavier_uniform_(self._h2.weight,
                                gain=nn.init.calculate_gain('relu'))
        nn.init.xavier_uniform_(self._h3.weight,
                                gain=nn.init.calculate_gain('linear'))
    

    @abstractmethod
    def forward(self, *args, **kwargs):
        pass



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/experiment_functions/01_run_experiment.ipynb.

# %% auto 0
__all__ = ['experiment_stepwise', 'experiment_stepwise_no_log', 'test', 'experiment_wandb', 'Weighted_Average',
           'experiment_wandb_optimize']

# %% ../../nbs/experiment_functions/01_run_experiment.ipynb 4
# MushroomRL:
from mushroom_rl.core import Core, Logger
from mushroom_rl.utils.dataset import compute_J,  parse_dataset

# Pytorch
import torch

# General libraries:
import numpy as np
import pickle
import os
from tqdm import tqdm, trange
import wandb

# %% ../../nbs/experiment_functions/01_run_experiment.ipynb 6
def experiment_stepwise(mdp,
                        run_params,
                        n_epochs,
                        n_steps,
                        n_episodes_test,
                        mdp_test = None,
                        rand_seed = None,
                        logger_dir = "./results/loggers",
                        results_dir = "./results",
                        dataset_log_freq = None):

    """
    TODO 1: Add function to (deep)save the entire agent to be able to reloead it later and potentially continue training.
    TODO 2: Provide functionality to turn of exploration during evaluation (e.g., epsilon=0 for DQN, sample_mean for SAC)

    Function to run an experiment with a given agent and mdp.
    Designed specifically for algorithms that learn on each step (e.g., SAC).

    Agents that have the fit function directly on data and are not trained
    stepwise via the Core class of Mushroom must have the attribute train_directly = True.

    The function assumes that that the mpd has a a demand attribute containing
    a list of historical demands. This is used as input for the directly trained agents.

    It will save the logger and save the logger containing the path. From the path it is 
    then possible to retrieve the results of J and the dataset.

    Args:
        mdp (object): MDP to be solved.
        run_params (dict): Dictionary with the agent and run number.
        n_epochs (int): Number of epochs to train the agent.
        n_steps (int): Number of steps per epoch.
        n_episodes_test (int): Number of episodes to evaluate the agent.
    """

    agent = run_params['agent']
    run = run_params['run']

    if rand_seed is not None:
        np.random.seed(rand_seed)
        torch.manual_seed(rand_seed)

    # Set-up logger
    logger = Logger(agent.name+f"_{run}", results_dir=f"{results_dir}/run_{run}")
    filename = f"{logger_dir}/logger_{agent.name}_{run}.pkl"
    os.makedirs(os.path.dirname(filename), exist_ok=True)
    pickle.dump(logger, open(f"{logger_dir}/logger_{agent.name}_{run}.pkl", "wb"))

    # ensure mdp starts from the same state
    mdp.reset()

    if mdp_test is not None:
        mdp_test.reset(state=0)

    # Start training

    logger.strong_line()
    logger.info('Experiment Algorithm: ' + agent.name)

    core = Core(agent, mdp)
    if mdp_test is not None:
        core_test = Core(agent, mdp_test)

    try:
        if agent.train_directly:
            agent.fit(mdp.demand)

    except:
   
        if mdp_test is not None:
            core_test = Core(agent, mdp_test)
        

        ### Initial evaluation
        if mdp_test is not None:
            dataset = core_test.evaluate(n_episodes=n_episodes_test, render=False, quiet = True)
        else:
            dataset = core.evaluate(n_episodes=n_episodes_test, render=False, quiet = True)
        s, *_ = parse_dataset(dataset)

        J = np.mean(compute_J(dataset, mdp.info.gamma))
        R = np.mean(compute_J(dataset))
        try:
            E = agent.policy.entropy(s)
        except:
            E = None

        logger.epoch_info(epoch=0, J=J, R=R, entropy=E)
        logger.log_numpy(epoch=0, J=J, R=R, entropy=E)
        logger.log_dataset(dataset, name_addition = "_epoch_0")

        ### Training
        initial_replay_size = agent._replay_memory._initial_size
        core.learn(n_steps=initial_replay_size, n_steps_per_fit=initial_replay_size, quiet=True)

        for n in trange(n_epochs, leave=False):
            core.learn(n_steps=n_steps, n_steps_per_fit=1, quiet = True)
            if mdp_test is not None:
                dataset = core_test.evaluate(n_episodes=n_episodes_test, render=False, quiet = True)
            else:
                dataset = core.evaluate(n_episodes=n_episodes_test, render=False, quiet = True)
            s, *_ = parse_dataset(dataset)

            J = np.mean(compute_J(dataset, mdp.info.gamma))
            R = np.mean(compute_J(dataset))
            try:
                E = agent.policy.entropy(s)
            except:
                E = None

            # logger.epoch_info(epoch = n+1, J=J, R=R, entropy=E)
            logger.log_numpy(epoch = n+1, J=J, R=R, entropy=E)

            if dataset_log_freq is not None:
                if n % dataset_log_freq == 0:
                    logger.log_dataset(dataset, name_addition = f"_epoch_{n+1}")
        
        ### Final evaluation
    
    if mdp_test is not None:
        dataset = core_test.evaluate(n_episodes=n_episodes_test, render=False, quiet=True)
    else:
        dataset = core.evaluate(n_episodes=n_episodes_test, render=False, quiet=True)
    
    s, *_ = parse_dataset(dataset)

    J = np.mean(compute_J(dataset, mdp.info.gamma))
    R = np.mean(compute_J(dataset))
    try:
        E = agent.policy.entropy(s)
    except:
        E = None

    logger.epoch_info(epoch = "final", J=J, R=R, E=E)
    logger.log_numpy(J=J, R=R, E=E)
    logger.log_dataset(dataset, name_addition = "_final")
        
    return logger

# %% ../../nbs/experiment_functions/01_run_experiment.ipynb 7
def experiment_stepwise_no_log(mdp,
                        agent,
                        n_epochs,
                        n_steps,
                        n_episodes_test,
                        mdp_test = None,
                        rand_seed = None,
                        dataset_log_freq = None):

    """
    TODO 1: Add function to (deep)save the entire agent to be able to reloead it later and potentially continue training.
    TODO 2: Provide functionalita to turn of exploration during evaluation (e.g., epsilon=0 for DQN, sample_mean for SAC)

    Function to run an experiment with a given agent and mdp.
    Designed specifically for algorithms that learn on each step (e.g., SAC) to tune hyperparameters.
    Does not use logger but simply returns the results.

    The function assumes that that the mpd has a a demand attribute containing
    a list of historical demands. This is used as input for the directly trained agents.

    Args:
        mdp (object): MDP to be solved.
        agent (object): Agent to solve the MDP.
        n_epochs (int): Number of epochs to train the agent.
        n_steps (int): Number of steps per epoch.
        n_episodes_test (int): Number of episodes to evaluate the agent.
    """

    if rand_seed is not None:
        np.random.seed(rand_seed)
        torch.manual_seed(rand_seed)

    # ensure mdp starts from the same state
    if mdp_test is not None:
        mdp_test.reset()
    else:
        mdp.reset()

    core = Core(agent, mdp)
    if mdp_test is not None:
        core_test = Core(agent, mdp_test)

    ### Training
    initial_replay_size = agent._replay_memory._initial_size
    core.learn(n_steps=initial_replay_size, n_steps_per_fit=initial_replay_size, quiet=True)

    R_hist = []

    for n in trange(n_epochs, leave=False):
        core.learn(n_steps=n_steps, n_steps_per_fit=1, quiet = True)
        if mdp_test is not None:
            dataset = core_test.evaluate(n_episodes=n_episodes_test, render=False, quiet = True)
        else:
            dataset = core.evaluate(n_episodes=n_episodes_test, render=False, quiet = True)

        R = np.mean(compute_J(dataset))

        R_hist.append(R)
        
    return R_hist

# %% ../../nbs/experiment_functions/01_run_experiment.ipynb 8
def test(agent, mdp_train, mdp_test, core_train, core_test, logger, n_episodes_test, n, log_wandb=True):

    print("testing")
    
    if mdp_test is not None:
        dataset = core_test.evaluate(n_episodes=n_episodes_test, render=False, quiet = True)
    else:
        dataset = core_train.evaluate(n_episodes=n_episodes_test, render=False, quiet = True)
    s, *_ = parse_dataset(dataset)

    J = np.mean(compute_J(dataset, mdp_train.info.gamma))
    R = np.mean(compute_J(dataset))
    
    print("check log wandb")

    try:
        E = agent.policy.entropy(s)
        if log_wandb:
            wandb.log({"test/J": J, "test/R": R, "test/entropy": E})
    except:
        E = None
        if log_wandb:
            wandb.log({"test/J": J, "test/R": R})
    print("wandb done")

    # logger.epoch_info(epoch = n+1, J=J, R=R, entropy=E)
    logger.log_numpy(epoch = n+1, J=J, R=R, entropy=E)

    print("testing done")
    
    return J, R


def experiment_wandb(   agent,
                        mdp_train,
                        mdp_test,
                        run_id,

                        n_epochs,
                        n_steps,
                        n_episodes_test,
                        rand_seed = None,
                        dataset_log_freq = None,
                        results_dir = "results",
                        ):

    """
    # TODO: update docstring
    """

    if rand_seed is not None:
        np.random.seed(rand_seed)
        torch.manual_seed(rand_seed)

    # Set-up logger
    path = f"./{results_dir}/{run_id}"
    os.makedirs(os.path.dirname(path), exist_ok=True)
    logger = Logger(agent.name+f"_{run_id}", results_dir=path)
    pickle.dump(logger, open(f"{path}/logger_{agent.name}_{run_id}.pkl", "wb"))

    # ensure mdp starts from the same state
    mdp_train.reset()

    if mdp_test is not None:
        mdp_test.reset(state=0)

    # Start training

    logger.strong_line()
    logger.info('Experiment Algorithm: ' + agent.name)

    core_train = Core(agent, mdp_train)
    if mdp_test is not None:
        core_test = Core(agent, mdp_test)

    ### Initial evaluation
    if mdp_test is not None:
        dataset = core_test.evaluate(n_episodes=n_episodes_test, render=False, quiet = True)
    else:
        dataset = core_train.evaluate(n_episodes=n_episodes_test, render=False, quiet = True)
    s, *_ = parse_dataset(dataset)

    J = np.mean(compute_J(dataset, mdp_train.info.gamma))
    R = np.mean(compute_J(dataset))
    wandb.log({"test/J": J, "test/R": R})
    
    try:
        E = agent.policy.entropy(s)
        wandb.log({"test/entropy": E})
    except:
        E = None

    logger.epoch_info(epoch=0, J=J, R=R, entropy=E)
    logger.log_numpy(epoch=0, J=J, R=R, entropy=E)
    logger.log_dataset(dataset, name_addition = "_epoch_0")

    artifact = wandb.Artifact("transitions_dataset", type='dataset')
    artifact.add_file(f"{path}/{agent.name}_{run_id}/dataset_epoch_0.pkl", name="dataset_epoch_0")

    try:   
        if agent.train_directly:
            if agent.train_mode == "direct":
                agent.fit(features=mdp_train.features, demand=mdp_train.demand)
                for i in trange(n_epochs):
                    if i==0:
                        J, R = test(agent, mdp_train, mdp_test, core_train, core_test, logger, n_episodes_test, i)
                    else:
                        wandb.log({"test/J": J, "test/R": R, "test/entropy": E})

            else:
                for i in trange(n_epochs):
                    agent.fit_epoch(mdp_train.features, mdp_train.demand)
                    agent.eval()
                    J, R = test(agent, mdp_train, mdp_test, core_train, core_test, logger, n_episodes_test, i)   
                    agent.train()
            
    except:
        ### Training
        initial_replay_size = agent._replay_memory._initial_size
        core_train.learn(n_steps=initial_replay_size, n_steps_per_fit=initial_replay_size, quiet=True)

        for n in trange(n_epochs, leave=False):
            core_train.learn(n_steps=n_steps, n_steps_per_fit=1, quiet = True)
            try:
                agent.policy.eval()
            except:
                pass
            if mdp_test is not None:
                dataset = core_test.evaluate(n_episodes=n_episodes_test, render=False, quiet = True)
            else:
                dataset = core_train.evaluate(n_episodes=n_episodes_test, render=False, quiet = True)
            s, *_ = parse_dataset(dataset)

            J = np.mean(compute_J(dataset, mdp_train.info.gamma))
            R = np.mean(compute_J(dataset))

            try:
                agent.policy.train()
            except:
                pass
            
            try:
                E = agent.policy.entropy(s)
                wandb.log({"test/J": J, "test/R": R, "test/entropy": E})
            except:
                E = None
                wandb.log({"test/J": J, "test/R": R})

            # logger.epoch_info(epoch = n+1, J=J, R=R, entropy=E)
            logger.log_numpy(epoch = n+1, J=J, R=R, entropy=E)

            if dataset_log_freq is not None:
                if (n+1) % dataset_log_freq == 0:
                    logger.log_dataset(dataset, name_addition = f"_epoch_{n+1}")
                    artifact.add_file(f"{path}/{agent.name}_{run_id}/dataset_epoch_{n+1}.pkl", name=f"dataset_epoch_{n+1}")
        
        ### Final evaluation
    
    try:
        agent.policy.eval()
    except:
        pass
    
    if mdp_test is not None:
        dataset = core_test.evaluate(n_episodes=n_episodes_test, render=False, quiet=True)
    else:
        dataset = core_train.evaluate(n_episodes=n_episodes_test, render=False, quiet=True)
    
    s, *_ = parse_dataset(dataset)

    J = np.mean(compute_J(dataset, mdp_train.info.gamma))
    R = np.mean(compute_J(dataset))
    wandb.log({"test/J": J, "test/R": R})
    try:
        E = agent.policy.entropy(s)
        wandb.log({"test/entropy": E})
    except:
        E = None

    logger.epoch_info(epoch = "final", J=J, R=R, E=E)
    logger.log_numpy(J=J, R=R, E=E)
    logger.log_dataset(dataset, name_addition = "_final")

    artifact.add_file(f"{path}/{agent.name}_{run_id}/dataset_final.pkl", name=f"dataset_final")

    wandb.log_artifact(artifact)
        
    return agent

# %% ../../nbs/experiment_functions/01_run_experiment.ipynb 9
class Weighted_Average:
    def __init__(self,
                 weight_final = 0.2,
                 start_tracking_epoch = 50, 
                 total_epochs = 200):
        self.weight_final = weight_final
        self.start_tracking_epoch = start_tracking_epoch
        self.totaol_epochs = total_epochs
        self.weight_datapoint = (1-weight_final)/(total_epochs-1-start_tracking_epoch)
        self.data = list()
        self.total_epochs = total_epochs

    def add_datapoint(self, datapoint, epoch):
        print("inside adding function")
        print("epoch", epoch)
        print("start_tracking_epoch", self.start_tracking_epoch)
        if epoch >= self.start_tracking_epoch:
            print("starting adding")
            print(epoch)
            print(self.total_epochs)
            if epoch < self.total_epochs-1:
                print("adding datapoint with weight")
                print("datapoing", datapoint)
                print("weight_datapoint", self.weight_datapoint)
                print("datapoint*weight_datapoint", datapoint*self.weight_datapoint)
                self.data.append(datapoint*self.weight_datapoint)
            elif epoch == self.total_epochs-1:
                print("adding datapoint with final weight")
                self.data.append(datapoint*self.weight_final)
            else:
                pass
        else:
            print("not adding datapoint")
        print("did")
    
    def get_result(self):
        print(self.data)
        return np.sum(self.data)

def experiment_wandb_optimize(   agent_l,
                        mdp_train_l,
                        mdp_val_l,
                        run_id,

                        weight_final,
                        start_tracking_epoch,

                        n_epochs,
                        n_steps,
                        n_episodes_test,
                        rand_seed = None,
                        dataset_log_freq = None,
                        results_dir = "results",
                        ):

    """
    # TODO: update docstring
    """


    if rand_seed is not None:
        np.random.seed(rand_seed)
        torch.manual_seed(rand_seed)

    weighted_j_list = list()

    print("starting loop")

    print("len mdp_train_l", len(mdp_train_l))

    for dataset_num in range(len(mdp_train_l)):

        print("dataset_num", dataset_num)


        mdp_train = mdp_train_l[dataset_num]
        mdp_val = mdp_val_l[dataset_num]
        agent = agent_l[dataset_num]

        print("setting up tracker")
        j_tracker = Weighted_Average(weight_final = weight_final, start_tracking_epoch = start_tracking_epoch, total_epochs = n_epochs)

        # Set-up logger
        print("setting up logger")
        path = f"./{results_dir}/{run_id}/_dataset_{dataset_num}"
        os.makedirs(os.path.dirname(path), exist_ok=True)
        logger = Logger(agent.name+f"_{run_id}_dataset_{dataset_num}", results_dir=path)
        pickle.dump(logger, open(f"{path}/logger_{agent.name}_{run_id}_dataset_{dataset_num}.pkl", "wb"))

        # ensure mdp starts from the same state
        mdp_train.reset()
        mdp_val.reset(state=0)

        # Start training

        print("start training")

        logger.strong_line()
        logger.info('Experiment Algorithm: ' + agent.name + f" on dataset {dataset_num}")

        core_train = Core(agent, mdp_train)
        core_val = Core(agent, mdp_val)

        print("going into training loop")

        try:   
            if agent.train_directly:
                if agent.train_mode == "direct":
                    agent.fit(features=mdp_train.features, demand=mdp_train.demand)
                    for i in trange(n_epochs):
                        if i==0:
                            J, R = test(agent, mdp_train, mdp_val, core_train, core_val, logger, n_episodes_test, i, log_wandb=False)
                        wandb.log({f"val/J_{dataset_num}": J, f"val/R_{dataset_num}": R})
                    print("did training of one model")

                else:
                    for i in trange(n_epochs):
                        agent.fit_epoch(mdp_train.features, mdp_train.demand)
                        print("done fitting")
                        J, R = test(agent, mdp_train, mdp_val, core_train, core_val, logger, n_episodes_test, i, log_wandb=False)
                        print("done testing")
                        wandb.log({f"val/J_{dataset_num}": J, f"val/R_{dataset_num}": R}) 
                        print(J)
                        j_tracker.add_datapoint(J, i)
                        print("added datapoint")
                
        except:
            ### Training
            initial_replay_size = agent._replay_memory._initial_size
            core_train.learn(n_steps=initial_replay_size, n_steps_per_fit=initial_replay_size, quiet=True)

            for n in trange(n_epochs, leave=False):
                core_train.learn(n_steps=n_steps, n_steps_per_fit=1, quiet = True)
                try:
                    agent.policy.eval()
                except:
                    pass
                dataset = core_val.evaluate(n_episodes=n_episodes_test, render=False, quiet = True)
                
                s, *_ = parse_dataset(dataset)

                J = np.mean(compute_J(dataset, mdp_train.info.gamma))
                R = np.mean(compute_J(dataset))

                j_tracker.add_datapoint(J, n)

                try:
                    agent.policy.train()
                except:
                    pass

                # logger.epoch_info(epoch = n+1, J=J, R=R, entropy=E)
                logger.log_numpy(epoch = n+1, J=J, R=R)
                wandb.log({f"val/J_{dataset_num}": J, f"val/R_{dataset_num}": R})
            
        weighted_j = j_tracker.get_result()
        print("weighted_j", weighted_j)

        weighted_j_list.append(weighted_j)
        

        
    ### Final evaluation
    print("ealculating final eval")
    print(weighted_j_list)
    weighed_j_total = np.mean(weighted_j_list)

    print("logging final eval to wandeb")
    wandb.log({f"val/J_weighted": weighed_j_total})
    
    return agent
